// use nix shell from 'nix'
// nix develop .#devShells.x86_64-linux.x86_64-w64-mingw32
// x86_64-w64-mingw32-g++ -luser32 main.cpp -o main.exe && ./main.exe
// or
// x86_64-w64-mingw32-g++ -luser32 main.cpp -o main.exe && cp main.exe /mnt/c/nix-hack

#include <filesystem>
#include <iostream>
#include <vector>
#include <optional>
#include <random>

#include <windows.h>

#include <gtest/gtest.h>

namespace fs = std::filesystem;

typedef std::string Path;
typedef std::string_view PathView;

void more();

fs::path canonPath(fs::path path, bool resolveSymlinks) {
    if (!path.is_absolute()) {
        std::cout << "not an absolute path" << std::endl;
    }
    if (fs::is_symlink(path) && !resolveSymlinks) {
        std::cout << "bailing out because path is a symlink while symlink resolving is disabled" << std::endl;
    }
    fs::path canonical_path;
    try {
        canonical_path = fs::weakly_canonical(path);
    }
    catch (std::filesystem::filesystem_error const &ex) {
        std::cout << "weakly_canonical path problem" << std::endl;
    }
    return canonical_path.make_preferred().lexically_normal();
}

fs::path absPath(fs::path path, std::optional<fs::path> dir = {}, bool resolveSymlinks = false)
{
    fs::path input_path(path);
    if (input_path.is_absolute()) {
        return canonPath(input_path, resolveSymlinks);
    } else {
        fs::path baseDir;
        if (!dir) {
            baseDir = std::filesystem::current_path();
        } else {
            baseDir = fs::path(*dir);  
        }
        if (!baseDir.is_absolute()) {
            std::cout << "dir is not an absolute path" << std::endl;
        }
        fs::path newDir = baseDir / input_path;
        return canonPath(newDir, resolveSymlinks);
    }
}

int main() {
    // FIXME windows native encoding
    // https://stackoverflow.com/questions/73631820/how-do-i-normalize-a-filepath-in-c-using-stdfilesystempath
    fs::path pp("/");
    std::cout << pp.is_absolute() << std::endl;  
    fs::path pq(".");
    std::cout << pq.is_absolute() << std::endl;    
    fs::path pq3("c:\\");
    std::cout << pq3.is_absolute() << std::endl;
    fs::path p("c:");
    std::cout << p.is_absolute() << std::endl;
    std::cout << p << " -> " << absPath(p) << std::endl;
    fs::path p1("c:\\");
    std::cout << p1 << " -> " << absPath(p1) << std::endl;
    fs::path p1j("c:\\nix\\..\\nix");
    std::cout << p1j << " -> " << absPath(p1j) << std::endl;
    std::cout << " da test -> " << absPath("foo", "c:\\nix-hack").string() << std::endl;
    fs::path p1ja("c:\\nix\\\\");
    std::cout << std::filesystem::canonical(p1ja).make_preferred().lexically_normal().string() << std::endl;
    fs::path p1jab("c:\\nix\\..\\nix\\asdf\\");
    std::cout << std::filesystem::weakly_canonical(p1jab).string() << std::endl;
    // NOTE: in general we want .string() so we have native encoding
    fs::path p1jab1(".");
    std::cout << std::filesystem::absolute(p1jab1).string() << std::endl;
    //more();

    // const std::filesystem::path cwd = std::filesystem::current_path();
    // std::cout << cwd << std::endl;
    // Path p99 = absPath("", Path(cwd.string()));
    // std::cout << p99 << std::endl;

    // FIXME paths larger than MAX_PATH
    // C:\Users\joschie\AppData\Local\Temp
    std::filesystem::path tmp = std::filesystem::temp_directory_path();
    std::filesystem::current_path(tmp);
    std::cout << "creating tmp path: " << tmp.string() << std::endl;
    auto d1 = tmp /
              "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890123456789012345678901234567890123456789012345678901234567890123456789"; // 200 chars
    std::filesystem::create_directories(d1);
    auto d2 = tmp /
              "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890123456789012345678901234567890abcdefgklmnopqrstuvwxyz12345678901234567890123456789012345678901234567890"; // 250 chars
    std::filesystem::create_directories(d2);
    auto d3 = tmp /
              "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz12345678abcdefgklmnopqrstuvwxyz12345678pqrstuvwxyz12345678901234567890123456789012345678901234567890123456789abcdefgklmnopqrstuvwxyz"; // 300 chars
    std::filesystem::create_directories(d3);

    // FIXME symlinks creation and recursive links resolving
    //   https://www.scivision.dev/softlink-in-windows-ntfs-drive/
    auto s1 = tmp / "nix-symlink-test";
    std::filesystem::create_directories(s1);
    std::filesystem::current_path(s1);

    // FIXME file symlink: in powershell.exe as admin:
    //    New-Item -Path C:\nix\troll.exe -ItemType SymbolicLink -Value C:\nix\nix.exe
    //    mklink LINK TARGET
    //    creates c:\foo -> c:\nix
    //    fs::create_symlink("main.exe", "foo.exe"); // FIXME this is not implemented for win10 so we use w32 instead in windows.h

    LPCSTR src = "c:\\nix\\nix.exe";
    LPCSTR dst = "c:\\nix\\main.exe";

    if (CreateSymbolicLinkA(dst, src, 0) == FALSE) {
        std::cout << "Error creating file symlink\n" << std::endl;
        DWORD error = ::GetLastError();
        std::string message = std::system_category().message(error);
        std::cout << message << std::endl;
    }

    // FIXME directory symlink: in powershell.exe as admin:
    //    New-Item -Path C:\foo -ItemType SymbolicLink -Value C:\nix
    //    mklink /d LINK TARGET
    //    creates c:\foo -> c:\nix
//    fs::create_directory_symlink(s1 / "foo", s1 / "bar");
//    fs::create_directory_symlink(s1 / "bar", s1 / "foo");

    LPCSTR srcd = "c:\\nix\\store";
    LPCSTR dstd = "c:\\nix\\lnk";

    if (CreateSymbolicLinkA(dstd, srcd, SYMBOLIC_LINK_FLAG_DIRECTORY) == FALSE) {
        std::cout << "Error creating directory symlink\n" << std::endl;
        DWORD error = ::GetLastError();
        std::string message = std::system_category().message(error);
        std::cout << message << std::endl;
    }

    LPCSTR srcd2 = "c:\\nix\\lnkd";
    LPCSTR dstd2 = "c:\\nix\\lnks";

    if (CreateSymbolicLinkA(dstd2, srcd2, SYMBOLIC_LINK_FLAG_DIRECTORY) == FALSE) {
        std::cout << "Error creating directory symlink\n" << std::endl;
        DWORD error = ::GetLastError();
        std::string message = std::system_category().message(error);
        std::cout << message << std::endl;
    }

    LPCSTR srcd3 = "c:\\nix\\lnks";
    LPCSTR dstd3 = "c:\\nix\\lnkd";

    if (CreateSymbolicLinkA(dstd3, srcd3, SYMBOLIC_LINK_FLAG_DIRECTORY) == FALSE) {
        std::cout << "Error creating directory symlink\n" << std::endl;
        DWORD error = ::GetLastError();
        std::string message = std::system_category().message(error);
        std::cout << message << std::endl;
    }

    Path zzz("c:\\nix\\lnks");
    std::cout << zzz << " -> " << absPath(zzz) << std::endl;
}

void more() {
    // Create a path object with an absolute path
    // std::filesystem::path absPath("c:\\foo\\bar");
    // std::filesystem::path absPath(R"(C:\\temp\\test.txt)");

    // if (absPath.is_absolute()) {
    //   std::cout << "The path is absolute: " << absPath << std::endl;
    // } else {
    //   std::cout << "The path is not absolute." << std::endl;
    // }

    std::vector <std::string> tests = {
            R"(C:\)",
            R"(C:\temp\test.txt)",
            R"(C:\\temp\\test.txt)",
            R"(\\?\C:\temp\test.txt)",
    };

    for (auto const &test: tests) {
        std::filesystem::path const p(test); // or boost::filesystem::path
        std::cout << std::endl << test << std::endl;
        std::cout << "root_name\t" << p.root_name().string() << std::endl;
        std::cout << "root_directory\t" << p.root_directory().string() << std::endl;
        std::cout << "root_path\t" << p.root_path().string() << std::endl;
        std::cout << "relative_path\t" << p.relative_path().string() << std::endl;
        std::cout << "is_absolute\t" << p.is_absolute() << std::endl;
        std::cout << "parent_path\t" << p.parent_path().string() << std::endl;
        std::cout << "filename\t" << p.filename().string() << std::endl;
        std::cout << "stem\t" << p.stem().string() << std::endl;
        std::cout << "extension\t" << p.extension().string() << std::endl;
    }

}
